package net.peachjean.modularwebapp;

import com.google.common.base.Function;
import com.google.common.base.Optional;
import com.google.common.collect.Iterables;
import com.google.inject.*;
import com.google.inject.servlet.GuiceFilter;
import org.eclipse.jetty.server.Connector;
import org.eclipse.jetty.server.DispatcherType;
import org.eclipse.jetty.server.Server;
import org.eclipse.jetty.server.handler.ContextHandlerCollection;
import org.eclipse.jetty.server.nio.SelectChannelConnector;
import org.eclipse.jetty.servlet.DefaultServlet;
import org.eclipse.jetty.servlet.FilterHolder;
import org.eclipse.jetty.servlet.ServletContextHandler;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.servlet.Filter;
import javax.servlet.ServletContext;
import javax.servlet.ServletContextListener;
import java.io.File;
import java.io.PrintStream;
import java.util.Collections;
import java.util.EnumSet;
import java.util.ServiceLoader;

public class ModularWebapp {

	private static final Logger logger = LoggerFactory.getLogger(ModularWebapp.class);

	private final int port;
	private final File resourceRoot;
	private final boolean production;
	private final Optional<PrintStream> serverOutput;
	private Server server;

	/**
	 *
	 * @param port the port on which the webapp will listen
	 * @param resourceRoot location of static resources for the webapp to serve
	 * @param production should we use the guice {@link Stage.PRODUCTION}?
	 * @param serverOutput optionally a {@link PrintStream} to output the server structure to
	 */
	public ModularWebapp(final int port, final File resourceRoot, final boolean production,
	                     final Optional<PrintStream> serverOutput) {
		this.port = port;
		this.resourceRoot = resourceRoot;
		this.production = production;
		this.serverOutput = serverOutput;
	}

	public void launchAndJoin() throws Exception {
		this.launch(true);
	}

	public void launch() throws Exception {
		this.launch(false);
	}

	private void launch(boolean waitForStop) throws Exception {
		// step 1 - Instantiate the server
		server = new Server();
		Connector connector = new SelectChannelConnector();
		connector.setPort(this.port);
		server.setConnectors(new Connector[]{connector});

		// step 2 - Setup the servlet context
		ContextHandlerCollection contexts = new ContextHandlerCollection();
		ServletContextHandler root = new ServletContextHandler(contexts, "/", ServletContextHandler.SESSIONS);
		root.setResourceBase(this.resourceRoot.getAbsolutePath());
		root.addServlet(DefaultServlet.class.getCanonicalName(), "/");

		// step 3 - instantiate guice modules
		Iterable<Module> modules = instantiateModules(root.getServletContext());

		// step 4 - instantiate injector
		Injector injector = Guice.createInjector(this.production ? Stage.PRODUCTION : Stage.DEVELOPMENT, modules);

		// step 5 - configure the servlet context with the guice filter
		root.addFilter(new FilterHolder(createGuiceFilter(injector)), "/*", EnumSet.allOf(DispatcherType.class));

		// step 6 - add ServletContextListeners to the servlet context so that they will participate in its lifecycle
		for (Key<?> key : injector.getBindings().keySet()) {
			if (ServletContextListener.class.isAssignableFrom(key.getTypeLiteral().getRawType())) {
				root.addEventListener(injector.getInstance((Key<ServletContextListener>) key));
			}
		}

		// step 7 - add contexts to server and start server
		server.setHandler(contexts);

		server.start();
		if (serverOutput.isPresent()) {
			serverOutput.get().println(server.dump());
		}
		if (waitForStop) {
			server.join();
		}
	}

	public void stop() throws Exception {
		if (server == null) {
			throw new IllegalArgumentException("Server has not been started.");
		}
		Server local = this.server;
		this.server = null;
		local.stop();
	}

	private Filter createGuiceFilter(final Injector injector) {
		return injector.getInstance(GuiceFilter.class);
	}

	protected Iterable<Module> instantiateModules(final ServletContext servletContext) {
		Iterable<ModularWebappExtension> extendsServices = locateExtensions();

		final Iterable<Module> modules = Iterables.concat(
				Iterables.transform(extendsServices, new ModularExtensionIterableFunction(servletContext)));

		if (logger.isDebugEnabled()) {
			logger.debug("Loading guice {} modules...", Iterables.size(modules));
			for (Module module : modules) {
				logger.debug("  Loading guice module {}.", module);
			}
		}
		return modules;
	}

	protected Iterable<ModularWebappExtension> locateExtensions() {
		return ServiceLoader.load(ModularWebappExtension.class);
	}

	private static class ModularExtensionIterableFunction implements Function<ModularWebappExtension, Iterable<Module>> {
		private final ServletContext servletContext;

		public ModularExtensionIterableFunction(final ServletContext servletContext) {
			this.servletContext = servletContext;
		}

		@Override
		public Iterable<Module> apply(ModularWebappExtension input) {
			return input != null
					? input.loadModules(servletContext)
					: Collections.<Module>emptyList();
		}
	}
}
